from __future__ import absolute_import, division, print_function

import os
import pkgutil
import subprocess
import sys
from unittest.mock import patch

import pytest

from distributed import Client, wait
from distributed.system import MEMORY_LIMIT
from distributed.utils_test import cleanup, loop, loop_in_thread, popen  # noqa: F401

from dask_cuda.utils import (
    get_cluster_configuration,
    get_device_total_memory,
    get_gpu_count_mig,
    get_gpu_uuid_from_index,
    get_n_gpus,
    wait_workers,
)


@patch.dict(os.environ, {"CUDA_VISIBLE_DEVICES": "0,3,7,8"})
def test_cuda_visible_devices_and_memory_limit_and_nthreads(loop):  # noqa: F811
    nthreads = 4
    with popen(["dask", "scheduler", "--port", "9359", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9359",
                "--host",
                "127.0.0.1",
                "--device-memory-limit",
                "1 MB",
                "--nthreads",
                str(nthreads),
                "--no-dashboard",
                "--worker-class",
                "dask_cuda.utils.MockWorker",
            ]
        ):
            with Client("127.0.0.1:9359", loop=loop) as client:
                assert wait_workers(client, n_gpus=4)

                def get_visible_devices():
                    return os.environ["CUDA_VISIBLE_DEVICES"]

                # verify 4 workers with the 4 expected CUDA_VISIBLE_DEVICES
                result = client.run(get_visible_devices)
                expected = {"0,3,7,8": 1, "3,7,8,0": 1, "7,8,0,3": 1, "8,0,3,7": 1}
                for v in result.values():
                    del expected[v]

                workers = client.scheduler_info()["workers"]
                for w in workers.values():
                    assert w["memory_limit"] == MEMORY_LIMIT // len(workers)

                assert len(expected) == 0


def test_rmm_pool(loop):  # noqa: F811
    rmm = pytest.importorskip("rmm")
    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9369",
                "--host",
                "127.0.0.1",
                "--rmm-pool-size",
                "2 GB",
                "--no-dashboard",
            ]
        ):
            with Client("127.0.0.1:9369", loop=loop) as client:
                assert wait_workers(client, n_gpus=get_n_gpus())

                memory_resource_type = client.run(
                    rmm.mr.get_current_device_resource_type
                )
                for v in memory_resource_type.values():
                    assert v is rmm.mr.PoolMemoryResource


def test_rmm_managed(loop):  # noqa: F811
    rmm = pytest.importorskip("rmm")
    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9369",
                "--host",
                "127.0.0.1",
                "--rmm-managed-memory",
                "--no-dashboard",
            ]
        ):
            with Client("127.0.0.1:9369", loop=loop) as client:
                assert wait_workers(client, n_gpus=get_n_gpus())

                memory_resource_type = client.run(
                    rmm.mr.get_current_device_resource_type
                )
                for v in memory_resource_type.values():
                    assert v is rmm.mr.ManagedMemoryResource


def test_rmm_async(loop):  # noqa: F811
    rmm = pytest.importorskip("rmm")

    driver_version = rmm._cuda.gpu.driverGetVersion()
    runtime_version = rmm._cuda.gpu.runtimeGetVersion()
    if driver_version < 11020 or runtime_version < 11020:
        pytest.skip("cudaMallocAsync not supported")

    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9369",
                "--host",
                "127.0.0.1",
                "--rmm-async",
                "--no-dashboard",
            ]
        ):
            with Client("127.0.0.1:9369", loop=loop) as client:
                assert wait_workers(client, n_gpus=get_n_gpus())

                memory_resource_type = client.run(
                    rmm.mr.get_current_device_resource_type
                )
                for v in memory_resource_type.values():
                    assert v is rmm.mr.CudaAsyncMemoryResource


def test_rmm_logging(loop):  # noqa: F811
    rmm = pytest.importorskip("rmm")
    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9369",
                "--host",
                "127.0.0.1",
                "--rmm-pool-size",
                "2 GB",
                "--rmm-log-directory",
                ".",
                "--no-dashboard",
            ]
        ):
            with Client("127.0.0.1:9369", loop=loop) as client:
                assert wait_workers(client, n_gpus=get_n_gpus())

                memory_resource_type = client.run(
                    rmm.mr.get_current_device_resource_type
                )
                for v in memory_resource_type.values():
                    assert v is rmm.mr.LoggingResourceAdaptor


@patch.dict(os.environ, {"CUDA_VISIBLE_DEVICES": "0"})
def test_dashboard_address(loop):  # noqa: F811
    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9369",
                "--dashboard-address",
                "127.0.0.1:9370",
            ]
        ):
            with Client("127.0.0.1:9369", loop=loop) as client:
                assert wait_workers(client, n_gpus=get_n_gpus())

                dashboard_addresses = client.run(
                    lambda dask_worker: dask_worker._dashboard_address
                )
                for v in dashboard_addresses.values():
                    assert v == "127.0.0.1:9370"


def test_unknown_argument():
    ret = subprocess.run(
        ["dask", "cuda", "worker", "--my-argument"], capture_output=True
    )
    assert ret.returncode != 0
    assert b"Scheduler address: --my-argument" in ret.stderr


@patch.dict(os.environ, {"CUDA_VISIBLE_DEVICES": "0"})
def test_pre_import(loop):  # noqa: F811
    module = None

    # Pick a module that isn't currently loaded
    for m in pkgutil.iter_modules():
        if m.ispkg and m.name not in sys.modules.keys():
            module = m.name
            break

    if module is None:
        pytest.skip("No module found that isn't already loaded")

    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9369",
                "--pre-import",
                module,
            ]
        ):
            with Client("127.0.0.1:9369", loop=loop) as client:
                assert wait_workers(client, n_gpus=get_n_gpus())

                imported = client.run(lambda: module in sys.modules)
                assert all(imported)


@pytest.mark.timeout(20)
@patch.dict(os.environ, {"CUDA_VISIBLE_DEVICES": "0"})
def test_pre_import_not_found():
    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        ret = subprocess.run(
            ["dask", "cuda", "worker", "127.0.0.1:9369", "--pre-import", "my_module"],
            capture_output=True,
        )
        assert ret.returncode != 0
        assert b"ModuleNotFoundError: No module named 'my_module'" in ret.stderr


def test_cuda_mig_visible_devices_and_memory_limit_and_nthreads(loop):  # noqa: F811
    uuids = get_gpu_count_mig(return_uuids=True)[1]
    # test only with some MIG Instances assuming the test bed
    # does not have a huge number of mig instances
    if len(uuids) > 0:
        cuda_visible_devices = ",".join([i.decode("utf-8") for i in uuids])
    else:
        pytest.skip("No MIG devices found")

    with patch.dict(os.environ, {"CUDA_VISIBLE_DEVICES": cuda_visible_devices}):
        nthreads = len(cuda_visible_devices)
        with popen(["dask", "scheduler", "--port", "9359", "--no-dashboard"]):
            with popen(
                [
                    "dask",
                    "cuda",
                    "worker",
                    "127.0.0.1:9359",
                    "--host",
                    "127.0.0.1",
                    "--nthreads",
                    str(nthreads),
                    "--no-dashboard",
                    "--worker-class",
                    "dask_cuda.utils.MockWorker",
                ]
            ):
                with Client("127.0.0.1:9359", loop=loop) as client:
                    assert wait_workers(client, n_gpus=len(uuids))
                    # Check to see if all workers are up and
                    # CUDA_VISIBLE_DEVICES cycles properly

                    def get_visible_devices():
                        return os.environ["CUDA_VISIBLE_DEVICES"]

                    result = client.run(get_visible_devices)
                    wait(result)
                    assert all(len(v.split(",")) == len(uuids) for v in result.values())
                    for i in range(len(uuids)):
                        assert set(
                            bytes(v.split(",")[i], "utf-8") for v in result.values()
                        ) == set(uuids)


def test_cuda_visible_devices_uuid(loop):  # noqa: F811
    gpu_uuid = get_gpu_uuid_from_index(0)

    with patch.dict(os.environ, {"CUDA_VISIBLE_DEVICES": gpu_uuid}):
        with popen(["dask", "scheduler", "--port", "9359", "--no-dashboard"]):
            with popen(
                [
                    "dask",
                    "cuda",
                    "worker",
                    "127.0.0.1:9359",
                    "--host",
                    "127.0.0.1",
                    "--no-dashboard",
                    "--worker-class",
                    "dask_cuda.utils.MockWorker",
                ]
            ):
                with Client("127.0.0.1:9359", loop=loop) as client:
                    assert wait_workers(client, n_gpus=1)

                    result = client.run(lambda: os.environ["CUDA_VISIBLE_DEVICES"])
                    assert list(result.values())[0] == gpu_uuid


def test_rmm_track_allocations(loop):  # noqa: F811
    rmm = pytest.importorskip("rmm")
    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9369",
                "--host",
                "127.0.0.1",
                "--rmm-pool-size",
                "2 GB",
                "--no-dashboard",
                "--rmm-track-allocations",
            ]
        ):
            with Client("127.0.0.1:9369", loop=loop) as client:
                assert wait_workers(client, n_gpus=get_n_gpus())

                memory_resource_type = client.run(
                    rmm.mr.get_current_device_resource_type
                )
                for v in memory_resource_type.values():
                    assert v is rmm.mr.TrackingResourceAdaptor

                memory_resource_upstream_type = client.run(
                    lambda: type(rmm.mr.get_current_device_resource().upstream_mr)
                )
                for v in memory_resource_upstream_type.values():
                    assert v is rmm.mr.PoolMemoryResource


@patch.dict(os.environ, {"CUDA_VISIBLE_DEVICES": "0"})
def test_get_cluster_configuration(loop):  # noqa: F811
    pytest.importorskip("rmm")
    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9369",
                "--host",
                "127.0.0.1",
                "--device-memory-limit",
                "30 B",
                "--rmm-pool-size",
                "2 GB",
                "--rmm-maximum-pool-size",
                "3 GB",
                "--no-dashboard",
                "--rmm-track-allocations",
            ]
        ):
            with Client("127.0.0.1:9369", loop=loop) as client:
                assert wait_workers(client, n_gpus=get_n_gpus())

                ret = get_cluster_configuration(client)
                wait(ret)
                assert ret["[plugin] RMMSetup"]["initial_pool_size"] == 2000000000
                assert ret["[plugin] RMMSetup"]["maximum_pool_size"] == 3000000000
                assert ret["jit-unspill"] is False
                assert ret["device-memory-limit"] == 30


@patch.dict(os.environ, {"CUDA_VISIBLE_DEVICES": "0"})
def test_worker_fraction_limits(loop):  # noqa: F811
    pytest.importorskip("rmm")
    with popen(["dask", "scheduler", "--port", "9369", "--no-dashboard"]):
        with popen(
            [
                "dask",
                "cuda",
                "worker",
                "127.0.0.1:9369",
                "--host",
                "127.0.0.1",
                "--device-memory-limit",
                "0.1",
                "--rmm-pool-size",
                "0.2",
                "--rmm-maximum-pool-size",
                "0.3",
                "--no-dashboard",
                "--rmm-track-allocations",
            ]
        ):
            with Client("127.0.0.1:9369", loop=loop) as client:
                assert wait_workers(client, n_gpus=get_n_gpus())

                device_total_memory = client.run(get_device_total_memory)
                wait(device_total_memory)
                _, device_total_memory = device_total_memory.popitem()

                ret = get_cluster_configuration(client)
                wait(ret)

                assert ret["device-memory-limit"] == int(device_total_memory * 0.1)
                assert (
                    ret["[plugin] RMMSetup"]["initial_pool_size"]
                    == (device_total_memory * 0.2) // 256 * 256
                )
                assert (
                    ret["[plugin] RMMSetup"]["maximum_pool_size"]
                    == (device_total_memory * 0.3) // 256 * 256
                )



from typing import Iterable
from unittest.mock import patch

import numpy as np
import pytest
from pandas.testing import assert_frame_equal

import dask
import dask.dataframe
from dask.dataframe.shuffle import shuffle_group
from dask.sizeof import sizeof
from dask.utils import format_bytes
from distributed import Client
from distributed.utils_test import gen_test
from distributed.worker import get_worker

import dask_cuda
import dask_cuda.proxify_device_objects
from dask_cuda.get_device_memory_objects import get_device_memory_ids
from dask_cuda.proxify_host_file import ProxifyHostFile
from dask_cuda.proxy_object import ProxyObject, asproxy, unproxy
from dask_cuda.utils import get_device_total_memory

cupy = pytest.importorskip("cupy")
cupy.cuda.set_allocator(None)
one_item_array = lambda: cupy.arange(1)
one_item_nbytes = one_item_array().nbytes

# While testing we don't want to unproxify `cupy.ndarray` even though
# it is on the incompatible_types list by default.
dask_cuda.proxify_device_objects.dispatch.dispatch(cupy.ndarray)
dask_cuda.proxify_device_objects.incompatible_types = ()  # type: ignore


@pytest.fixture(scope="module")
def root_dir(tmp_path_factory):
    tmpdir = tmp_path_factory.mktemp("jit-unspill")
    # Make the "disk" serializer available and use a tmp directory
    if ProxifyHostFile._spill_to_disk is None:
        ProxifyHostFile(
            worker_local_directory=tmpdir.name,
            device_memory_limit=1024,
            memory_limit=1024,
        )
    assert ProxifyHostFile._spill_to_disk is not None

    # In order to use the same tmp dir, we use `root_dir` for all
    # ProxifyHostFile creations. Notice, we use `..` to remove the
    # `jit-unspill-disk-storage` part added by the
    # ProxifyHostFile implicitly.
    return str(ProxifyHostFile._spill_to_disk.root_dir / "..")


def is_proxies_equal(p1: Iterable[ProxyObject], p2: Iterable[ProxyObject]):
    """Check that two collections of proxies contains the same proxies (unordered)

    In order to avoid deserializing proxy objects when comparing them,
    this function compares object IDs.
    """

    ids1 = sorted([id(p) for p in p1])
    ids2 = sorted([id(p) for p in p2])
    return ids1 == ids2


def test_one_dev_item_limit(root_dir):
    dhf = ProxifyHostFile(
        worker_local_directory=root_dir,
        device_memory_limit=one_item_nbytes,
        memory_limit=1000,
    )

    a1 = one_item_array() + 42
    a2 = one_item_array()
    dhf["k1"] = a1
    dhf["k2"] = a2
    dhf.manager.validate()

    # Check k1 is spilled because of the newer k2
    k1 = dhf["k1"]
    k2 = dhf["k2"]
    assert k1._pxy_get().is_serialized()
    assert not k2._pxy_get().is_serialized()
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k2])

    # Accessing k1 spills k2 and unspill k1
    k1_val = k1[0]
    assert k1_val == 42
    assert k2._pxy_get().is_serialized()
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k2])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k1])

    # Duplicate arrays changes nothing
    dhf["k3"] = [k1, k2]
    assert not k1._pxy_get().is_serialized()
    assert k2._pxy_get().is_serialized()
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k2])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k1])

    # Adding a new array spills k1 and k2
    dhf["k4"] = one_item_array()
    k4 = dhf["k4"]
    assert k1._pxy_get().is_serialized()
    assert k2._pxy_get().is_serialized()
    assert not dhf["k4"]._pxy_get().is_serialized()
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1, k2])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k4])

    # Accessing k2 spills k1 and k4
    k2[0]
    assert k1._pxy_get().is_serialized()
    assert dhf["k4"]._pxy_get().is_serialized()
    assert not k2._pxy_get().is_serialized()
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1, k4])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k2])

    # Deleting k2 does not change anything since k3 still holds a
    # reference to the underlying proxy object
    assert dhf.manager._dev.mem_usage() == one_item_nbytes
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1, k4])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k2])
    del dhf["k2"]
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1, k4])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k2])

    # Overwriting k3 with a non-cuda object and deleting k2
    # should empty the device
    dhf["k3"] = "non-cuda-object"
    del k2
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1, k4])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [])

    # Adding the underlying proxied of k1 doesn't change anything.
    # The host file detects that k1_ary is already proxied by the
    # existing proxy object k1.
    k1_ary = unproxy(k1)
    dhf["k5"] = k1_ary
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k4])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k1])

    # Clean up
    del k1, k4
    dhf.clear()
    assert len(dhf.manager) == 0


def test_one_item_host_limit(capsys, root_dir):
    memory_limit = sizeof(asproxy(one_item_array(), serializers=("dask", "pickle")))
    dhf = ProxifyHostFile(
        worker_local_directory=root_dir,
        device_memory_limit=one_item_nbytes,
        memory_limit=memory_limit,
    )

    a1 = one_item_array() + 1
    a2 = one_item_array() + 2
    dhf["k1"] = a1
    dhf["k2"] = a2
    dhf.manager.validate()

    # Check k1 is spilled because of the newer k2
    k1 = dhf["k1"]
    k2 = dhf["k2"]
    assert k1._pxy_get().is_serialized()
    assert not k2._pxy_get().is_serialized()
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._disk.get_proxies(), [])
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k2])

    # Check k1 is spilled to disk and k2 is spilled to host
    dhf["k3"] = one_item_array() + 3
    k3 = dhf["k3"]
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._disk.get_proxies(), [k1])
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k2])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k3])
    dhf.manager.validate()

    # Accessing k2 spills k3 and unspill k2
    k2_val = k2[0]
    assert k2_val == 2
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._disk.get_proxies(), [k1])
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k3])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k2])

    # Adding a new array spill k3 to disk and k2 to host
    dhf["k4"] = one_item_array() + 4
    k4 = dhf["k4"]
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._disk.get_proxies(), [k1, k3])
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k2])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k4])

    # Accessing k1 unspills k1 directly to device and spills k4 to host
    k1_val = k1[0]
    assert k1_val == 1
    dhf.manager.validate()
    assert is_proxies_equal(dhf.manager._disk.get_proxies(), [k2, k3])
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k4])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k1])

    # Clean up
    del k1, k2, k3, k4
    dhf.clear()
    assert len(dhf.manager) == 0


def test_spill_on_demand(root_dir):
    """
    Test spilling on demand by disabling the device_memory_limit
    and allocating two large buffers that will otherwise fail because
    of spilling on demand.
    """
    rmm = pytest.importorskip("rmm")
    if not hasattr(rmm.mr, "FailureCallbackResourceAdaptor"):
        pytest.skip("RMM doesn't implement FailureCallbackResourceAdaptor")

    total_mem = get_device_total_memory()
    dhf = ProxifyHostFile(
        worker_local_directory=root_dir,
        device_memory_limit=2 * total_mem,
        memory_limit=2 * total_mem,
        spill_on_demand=True,
    )
    for i in range(2):
        dhf[i] = rmm.DeviceBuffer(size=total_mem // 2 + 1)


@pytest.mark.parametrize("jit_unspill", [True, False])
@gen_test(timeout=20)
async def test_local_cuda_cluster(jit_unspill):
    """Testing spilling of a proxied cudf dataframe in a local cuda cluster"""
    cudf = pytest.importorskip("cudf")
    dask_cudf = pytest.importorskip("dask_cudf")

    def task(x):
        assert isinstance(x, cudf.DataFrame)
        if jit_unspill:
            # Check that `x` is a proxy object and the proxied DataFrame is serialized
            assert "ProxyObject" in str(type(x))
            assert x._pxy_get().serializer == "dask"
        else:
            assert type(x) == cudf.DataFrame
        assert len(x) == 10  # Trigger deserialization
        return x

    # Notice, setting `device_memory_limit=1B` to trigger spilling
    async with dask_cuda.LocalCUDACluster(
        n_workers=1,
        device_memory_limit="1B",
        jit_unspill=jit_unspill,
        asynchronous=True,
    ) as cluster:
        async with Client(cluster, asynchronous=True) as client:
            df = cudf.DataFrame({"a": range(10)})
            ddf = dask_cudf.from_cudf(df, npartitions=1)
            ddf = ddf.map_partitions(task, meta=df.head())
            got = await client.compute(ddf)
            assert_frame_equal(got.to_pandas(), df.to_pandas())


def test_dataframes_share_dev_mem(root_dir):
    cudf = pytest.importorskip("cudf")

    df = cudf.DataFrame({"a": range(10)})
    grouped = shuffle_group(df, "a", 0, 2, 2, False, 2)
    view1 = grouped[0]
    view2 = grouped[1]
    # Even though the two dataframe doesn't point to the same cudf.Buffer object
    assert view1["a"].data is not view2["a"].data
    # They still share the same underlying device memory
    view1["a"].data.ptr == view2["a"].data.ptr

    dhf = ProxifyHostFile(
        worker_local_directory=root_dir, device_memory_limit=160, memory_limit=1000
    )
    dhf["v1"] = view1
    dhf["v2"] = view2
    v1 = dhf["v1"]
    v2 = dhf["v2"]
    # The device_memory_limit is not exceeded since both dataframes share device memory
    assert not v1._pxy_get().is_serialized()
    assert not v2._pxy_get().is_serialized()
    # Now the device_memory_limit is exceeded, which should evict both dataframes
    dhf["k1"] = one_item_array()
    assert v1._pxy_get().is_serialized()
    assert v2._pxy_get().is_serialized()


def test_cudf_get_device_memory_objects():
    cudf = pytest.importorskip("cudf")
    objects = [
        cudf.DataFrame({"a": range(10), "b": range(10)}, index=reversed(range(10))),
        cudf.MultiIndex(
            levels=[[1, 2], ["blue", "red"]], codes=[[0, 0, 1, 1], [1, 0, 1, 0]]
        ),
    ]
    res = get_device_memory_ids(objects)
    assert len(res) == 4, "We expect four buffer objects"


def test_externals(root_dir):
    """Test adding objects directly to the manager

    Add an object directly to the manager makes it count against the
    device_memory_limit but isn't part of the store.

    Normally, we use __setitem__ to store objects in the hostfile and make it
    count against the device_memory_limit with the inherent consequence that
    the objects are not freeable before subsequential calls to __delitem__.
    This is a problem for long running tasks that want objects to count against
    the device_memory_limit while freeing them ASAP without explicit calls to
    __delitem__.
    """
    dhf = ProxifyHostFile(
        worker_local_directory=root_dir,
        device_memory_limit=one_item_nbytes,
        memory_limit=1000,
    )
    dhf["k1"] = one_item_array()
    k1 = dhf["k1"]
    k2, incompatible_type_found = dhf.manager.proxify(one_item_array())
    assert not incompatible_type_found
    # `k2` isn't part of the store but still triggers spilling of `k1`
    assert len(dhf) == 1
    assert k1._pxy_get().is_serialized()
    assert not k2._pxy_get().is_serialized()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k2])
    assert dhf.manager._dev._mem_usage == one_item_nbytes

    k1[0]  # Trigger spilling of `k2`
    assert not k1._pxy_get().is_serialized()
    assert k2._pxy_get().is_serialized()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k2])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k1])
    assert dhf.manager._dev._mem_usage == one_item_nbytes

    k2[0]  # Trigger spilling of `k1`
    assert k1._pxy_get().is_serialized()
    assert not k2._pxy_get().is_serialized()
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [k2])
    assert dhf.manager._dev._mem_usage == one_item_nbytes

    # Removing `k2` also removes it from the tally
    del k2
    assert is_proxies_equal(dhf.manager._host.get_proxies(), [k1])
    assert is_proxies_equal(dhf.manager._dev.get_proxies(), [])
    assert dhf.manager._dev._mem_usage == 0


@patch("dask_cuda.proxify_device_objects.incompatible_types", (cupy.ndarray,))
def test_incompatible_types(root_dir):
    """Check that ProxifyHostFile unproxifies `cupy.ndarray` on retrieval

    Notice, in this test we add `cupy.ndarray` to the incompatible_types temporarily.
    """
    cupy = pytest.importorskip("cupy")
    cudf = pytest.importorskip("cudf")
    dhf = ProxifyHostFile(
        worker_local_directory=root_dir, device_memory_limit=100, memory_limit=100
    )

    # We expect `dhf` to unproxify `a1` (but not `a2`) on retrieval
    a1, a2 = (cupy.arange(9), cudf.Series([1, 2, 3]))
    dhf["a"] = (a1, a2)
    b1, b2 = dhf["a"]
    assert a1 is b1
    assert isinstance(b2, ProxyObject)
    assert a2 is unproxy(b2)


@pytest.mark.parametrize("npartitions", [1, 2, 3])
@pytest.mark.parametrize("compatibility_mode", [True, False])
@gen_test(timeout=20)
async def test_compatibility_mode_dataframe_shuffle(compatibility_mode, npartitions):
    cudf = pytest.importorskip("cudf")

    def is_proxy_object(x):
        return "ProxyObject" in str(type(x))

    with dask.config.set(jit_unspill_compatibility_mode=compatibility_mode):
        async with dask_cuda.LocalCUDACluster(
            n_workers=1, jit_unspill=True, asynchronous=True
        ) as cluster:
            async with Client(cluster, asynchronous=True) as client:
                ddf = dask.dataframe.from_pandas(
                    cudf.DataFrame({"key": np.arange(10)}), npartitions=npartitions
                )
                res = ddf.shuffle(on="key", shuffle="tasks").persist()

                # With compatibility mode on, we shouldn't encounter any proxy objects
                if compatibility_mode:
                    assert "ProxyObject" not in str(type(await client.compute(res)))
                res = await client.compute(res.map_partitions(is_proxy_object))
                res = res.to_list()

                if compatibility_mode:
                    assert not any(res)  # No proxy objects
                else:
                    assert all(res)  # Only proxy objects


@gen_test(timeout=60)
async def test_worker_force_spill_to_disk():
    """Test Dask triggering CPU-to-Disk spilling"""
    cudf = pytest.importorskip("cudf")

    with dask.config.set({"distributed.worker.memory.terminate": False}):
        async with dask_cuda.LocalCUDACluster(
            n_workers=1, device_memory_limit="1MB", jit_unspill=True, asynchronous=True
        ) as cluster:
            async with Client(cluster, asynchronous=True) as client:
                # Create a df that are spilled to host memory immediately
                df = cudf.DataFrame({"key": np.arange(10**8)})
                ddf = dask.dataframe.from_pandas(df, npartitions=1).persist()
                await ddf

                async def f():
                    """Trigger a memory_monitor() and reset memory_limit"""
                    w = get_worker()
                    # Set a host memory limit that triggers spilling to disk
                    w.memory_manager.memory_pause_fraction = False
                    memory = w.monitor.proc.memory_info().rss
                    w.memory_manager.memory_limit = memory - 10**8
                    w.memory_manager.memory_target_fraction = 1
                    print(w.memory_manager.data)
                    await w.memory_manager.memory_monitor(w)
                    # Check that host memory are freed
                    assert w.monitor.proc.memory_info().rss < memory - 10**7
                    w.memory_manager.memory_limit = memory * 10  # Un-limit

                await client.submit(f)
                log = str(await client.get_worker_logs())
                # Check that the worker doesn't complain about unmanaged memory
                assert "Unmanaged memory use is high" not in log


def test_on_demand_debug_info():
    """Test worker logging when on-demand-spilling fails"""
    rmm = pytest.importorskip("rmm")
    if not hasattr(rmm.mr, "FailureCallbackResourceAdaptor"):
        pytest.skip("RMM doesn't implement FailureCallbackResourceAdaptor")

    rmm_pool_size = 2**20

    def task():
        (
            rmm.DeviceBuffer(size=rmm_pool_size // 2),
            rmm.DeviceBuffer(size=rmm_pool_size // 2),
            rmm.DeviceBuffer(size=rmm_pool_size),  # Trigger OOM
        )

    with dask_cuda.LocalCUDACluster(
        n_workers=1,
        jit_unspill=True,
        rmm_pool_size=rmm_pool_size,
        rmm_maximum_pool_size=rmm_pool_size,
        rmm_track_allocations=True,
    ) as cluster:
        with Client(cluster) as client:
            # Warmup, which trigger the initialization of spill on demand
            client.submit(range, 10).result()

            # Submit too large RMM buffer
            with pytest.raises(MemoryError, match="Maximum pool size exceeded"):
                client.submit(task).result()

            log = str(client.get_worker_logs())
            size = format_bytes(rmm_pool_size)
            assert f"WARNING - RMM allocation of {size} failed" in log
            assert f"RMM allocs: {size}" in log
            assert "traceback:" in log



import contextlib
import math
from collections import ChainMap
from time import perf_counter

import numpy as np
import pandas as pd

import dask
from dask.base import tokenize
from dask.dataframe.core import new_dd_object
from dask.distributed import performance_report, wait
from dask.utils import format_bytes, parse_bytes

from dask_cuda.benchmarks.common import Config, execute_benchmark
from dask_cuda.benchmarks.utils import (
    as_noop,
    parse_benchmark_args,
    print_key_value,
    print_separator,
    print_throughput_bandwidth,
)

# Benchmarking cuDF merge operation based on
# <https://gist.github.com/rjzamora/0ffc35c19b5180ab04bbf7c793c45955>


def generate_chunk(i_chunk, local_size, num_chunks, chunk_type, frac_match, gpu):
    # Setting a seed that triggers max amount of comm in the two-GPU case.
    if gpu:
        import cupy as xp

        import cudf as xdf
    else:
        import numpy as xp
        import pandas as xdf

    xp.random.seed(2**32 - 1)

    chunk_type = chunk_type or "build"
    frac_match = frac_match or 1.0
    if chunk_type == "build":
        # Build dataframe
        #
        # "key" column is a unique sample within [0, local_size * num_chunks)
        #
        # "shuffle" column is a random selection of partitions (used for shuffle)
        #
        # "payload" column is a random permutation of the chunk_size

        start = local_size * i_chunk
        stop = start + local_size

        parts_array = xp.arange(num_chunks, dtype="int64")
        suffle_array = xp.repeat(parts_array, math.ceil(local_size / num_chunks))

        df = xdf.DataFrame(
            {
                "key": xp.arange(start, stop=stop, dtype="int64"),
                "shuffle": xp.random.permutation(suffle_array)[:local_size],
                "payload": xp.random.permutation(xp.arange(local_size, dtype="int64")),
            }
        )
    else:
        # Other dataframe
        #
        # "key" column matches values from the build dataframe
        # for a fraction (`frac_match`) of the entries. The matching
        # entries are perfectly balanced across each partition of the
        # "base" dataframe.
        #
        # "payload" column is a random permutation of the chunk_size

        # Step 1. Choose values that DO match
        sub_local_size = local_size // num_chunks
        sub_local_size_use = max(int(sub_local_size * frac_match), 1)
        arrays = []
        for i in range(num_chunks):
            bgn = (local_size * i) + (sub_local_size * i_chunk)
            end = bgn + sub_local_size
            ar = xp.arange(bgn, stop=end, dtype="int64")
            arrays.append(xp.random.permutation(ar)[:sub_local_size_use])
        key_array_match = xp.concatenate(tuple(arrays), axis=0)

        # Step 2. Add values that DON'T match
        missing_size = local_size - key_array_match.shape[0]
        start = local_size * num_chunks + local_size * i_chunk
        stop = start + missing_size
        key_array_no_match = xp.arange(start, stop=stop, dtype="int64")

        # Step 3. Combine and create the final dataframe chunk (dask_cudf partition)
        key_array_combine = xp.concatenate(
            (key_array_match, key_array_no_match), axis=0
        )
        df = xdf.DataFrame(
            {
                "key": xp.random.permutation(key_array_combine),
                "payload": xp.random.permutation(xp.arange(local_size, dtype="int64")),
            }
        )
    return df


def get_random_ddf(chunk_size, num_chunks, frac_match, chunk_type, args):

    parts = [chunk_size for _ in range(num_chunks)]
    device_type = True if args.type == "gpu" else False
    meta = generate_chunk(0, 4, 1, chunk_type, None, device_type)
    divisions = [None] * (len(parts) + 1)

    name = "generate-data-" + tokenize(chunk_size, num_chunks, frac_match, chunk_type)

    graph = {
        (name, i): (
            generate_chunk,
            i,
            part,
            len(parts),
            chunk_type,
            frac_match,
            device_type,
        )
        for i, part in enumerate(parts)
    }

    ddf = new_dd_object(graph, name, meta, divisions)

    if chunk_type == "build":
        if not args.no_shuffle:
            divisions = [i for i in range(num_chunks)] + [num_chunks]
            return ddf.set_index("shuffle", divisions=tuple(divisions))
        else:
            del ddf["shuffle"]

    return ddf


def merge(args, ddf1, ddf2):

    # Allow default broadcast behavior, unless
    # "--shuffle-join" or "--broadcast-join" was
    # specified (with "--shuffle-join" taking
    # precedence)
    broadcast = False if args.shuffle_join else (True if args.broadcast_join else None)

    # The merge/join operation
    ddf_join = ddf1.merge(ddf2, on=["key"], how="inner", broadcast=broadcast)
    if args.set_index:
        ddf_join = ddf_join.set_index("key")
    if args.backend == "dask-noop":
        t1 = perf_counter()
        ddf_join = as_noop(ddf_join)
        noopify_duration = perf_counter() - t1
    else:
        noopify_duration = 0
    wait(ddf_join.persist())
    return noopify_duration


def bench_once(client, args, write_profile=None):
    # Generate random Dask dataframes
    n_workers = len(client.scheduler_info()["workers"])
    # Allow the number of chunks to vary between
    # the "base" and "other" DataFrames
    args.base_chunks = args.base_chunks or n_workers
    args.other_chunks = args.other_chunks or n_workers
    ddf_base = get_random_ddf(
        args.chunk_size, args.base_chunks, args.frac_match, "build", args
    ).persist()
    ddf_other = get_random_ddf(
        args.chunk_size, args.other_chunks, args.frac_match, "other", args
    ).persist()
    wait(ddf_base)
    wait(ddf_other)

    assert len(ddf_base.dtypes) == 2
    assert len(ddf_other.dtypes) == 2
    data_processed = len(ddf_base) * sum([t.itemsize for t in ddf_base.dtypes])
    data_processed += len(ddf_other) * sum([t.itemsize for t in ddf_other.dtypes])

    # Get contexts to use (defaults to null contexts that doesn't do anything)
    ctx1, ctx2 = contextlib.nullcontext(), contextlib.nullcontext()
    if args.backend == "explicit-comms":
        ctx1 = dask.config.set(explicit_comms=True)
    if write_profile is not None:
        ctx2 = performance_report(filename=args.profile)

    with ctx1:
        with ctx2:
            t1 = perf_counter()
            noopify_duration = merge(args, ddf_base, ddf_other)
            duration = perf_counter() - t1 - noopify_duration

    return (data_processed, duration)


def pretty_print_results(args, address_to_index, p2p_bw, results):
    broadcast = (
        False if args.shuffle_join else (True if args.broadcast_join else "default")
    )

    if args.markdown:
        print("```")
    print("Merge benchmark")
    print_separator(separator="-")
    print_key_value(key="Backend", value=f"{args.backend}")
    print_key_value(key="Merge type", value=f"{args.type}")
    print_key_value(key="Rows-per-chunk", value=f"{args.chunk_size}")
    print_key_value(key="Base-chunks", value=f"{args.base_chunks}")
    print_key_value(key="Other-chunks", value=f"{args.other_chunks}")
    print_key_value(key="Broadcast", value=f"{broadcast}")
    print_key_value(key="Protocol", value=f"{args.protocol}")
    print_key_value(key="Device(s)", value=f"{args.devs}")
    if args.device_memory_limit:
        print_key_value(
            key="Device memory limit", value=f"{format_bytes(args.device_memory_limit)}"
        )
    print_key_value(key="RMM Pool", value=f"{not args.disable_rmm_pool}")
    print_key_value(key="Frac-match", value=f"{args.frac_match}")
    if args.protocol == "ucx":
        print_key_value(key="TCP", value=f"{args.enable_tcp_over_ucx}")
        print_key_value(key="InfiniBand", value=f"{args.enable_infiniband}")
        print_key_value(key="NVLink", value=f"{args.enable_nvlink}")
    print_key_value(key="Worker thread(s)", value=f"{args.threads_per_worker}")
    print_key_value(key="Data processed", value=f"{format_bytes(results[0][0])}")
    if args.markdown:
        print("\n```")

    data_processed, durations = zip(*results)
    print_throughput_bandwidth(
        args, durations, data_processed, p2p_bw, address_to_index
    )


def create_tidy_results(
    args,
    p2p_bw: np.ndarray,
    results,
):
    broadcast = (
        False if args.shuffle_join else (True if args.broadcast_join else "default")
    )
    configuration = {
        "dataframe_type": "cudf" if args.type == "gpu" else "pandas",
        "backend": args.backend,
        "merge_type": args.type,
        "base_chunks": args.base_chunks,
        "other_chunks": args.other_chunks,
        "broadcast": broadcast,
        "rows_per_chunk": args.chunk_size,
        "ignore_size": args.ignore_size,
        "frac_match": args.frac_match,
        "devices": args.devs,
        "device_memory_limit": args.device_memory_limit,
        "worker_threads": args.threads_per_worker,
        "rmm_pool": not args.disable_rmm_pool,
        "protocol": args.protocol,
        "tcp": args.enable_tcp_over_ucx,
        "ib": args.enable_infiniband,
        "nvlink": args.enable_nvlink,
        "nreps": args.runs,
    }
    timing_data = pd.DataFrame(
        [
            pd.Series(
                data=ChainMap(
                    configuration,
                    {"wallclock": duration, "data_processed": data_processed},
                )
            )
            for data_processed, duration in results
        ]
    )
    return timing_data, p2p_bw


def parse_args():
    special_args = [
        {
            "name": [
                "-b",
                "--backend",
            ],
            "choices": ["dask", "explicit-comms", "dask-noop"],
            "default": "dask",
            "type": str,
            "help": "The backend to use.",
        },
        {
            "name": [
                "-t",
                "--type",
            ],
            "choices": ["cpu", "gpu"],
            "default": "gpu",
            "type": str,
            "help": "Do merge with GPU or CPU dataframes",
        },
        {
            "name": [
                "-c",
                "--chunk-size",
            ],
            "default": 1_000_000,
            "metavar": "n",
            "type": int,
            "help": "Chunk size (default 1_000_000)",
        },
        {
            "name": "--base-chunks",
            "default": None,
            "type": int,
            "help": "Number of base-DataFrame partitions (default: n_workers)",
        },
        {
            "name": "--other-chunks",
            "default": None,
            "type": int,
            "help": "Number of other-DataFrame partitions (default: n_workers)",
        },
        {
            "name": "--broadcast-join",
            "action": "store_true",
            "help": "Use broadcast join when possible.",
        },
        {
            "name": "--shuffle-join",
            "action": "store_true",
            "help": "Use shuffle join (takes precedence over '--broadcast-join').",
        },
        {
            "name": "--ignore-size",
            "default": "1 MiB",
            "metavar": "nbytes",
            "type": parse_bytes,
            "help": "Ignore messages smaller than this (default '1 MB')",
        },
        {
            "name": "--frac-match",
            "default": 0.3,
            "type": float,
            "help": "Fraction of rows that matches (default 0.3)",
        },
        {
            "name": "--no-shuffle",
            "action": "store_true",
            "help": "Don't shuffle the keys of the left (base) dataframe.",
        },
        {
            "name": "--runs",
            "default": 3,
            "type": int,
            "help": "Number of runs",
        },
        {
            "name": [
                "-s",
                "--set-index",
            ],
            "action": "store_true",
            "help": "Call set_index on the key column to sort the joined dataframe.",
        },
    ]

    return parse_benchmark_args(
        description="Distributed merge (dask/cudf) benchmark", args_list=special_args
    )


if __name__ == "__main__":
    execute_benchmark(
        Config(
            args=parse_args(),
            bench_once=bench_once,
            create_tidy_results=create_tidy_results,
            pretty_print_results=pretty_print_results,
        )
    )
